# This file serves to test the operator.py module
from Dotua.operator import Operator as op
from Dotua.autodiff import AutoDiff as ad
import numpy as np

x, z, n = tuple(ad.create_scalar([0, 1, 2048]))
v = ad.create_vector([[.6,.25,.5,.75]])[0]
v2 = op.sin(v)
orig_vals = np.copy(v._val)
y = 0


def test_sin():
    # Scalar
    sx = op.sin(x)
    assert np.sin(x._val) == sx._val
    for k in x._jacobian.keys():
        assert sx.partial(k) == x.partial(k) * np.cos(x._val)

    # Simple Vector Values
    sv = op.sin(v)
    assert ~(np.sin(orig_vals) - sv._val).any()

    # Simple Vector Jacobian
    for key in sv._dict.keys():
        assert ~(sv._dict[key] - v.getDerivative(key) * np.cos(v._val) * np.eye(len(v._val))).any()

    # Complex Vector Values
    scv = op.sin(sv)

    assert ~(np.sin(np.sin(orig_vals)) - scv._val).any()

    # Complex Vector Jacobian
    for key in scv._dict.keys():
        assert ~(scv._dict[key] - sv.getDerivative(key) * np.cos(sv._val) * np.eye(len(sv._val))).any()

    # Constant
    assert op.sin(y) == np.sin(y)


def test_cos():
    # Scalar
    cx = op.cos(x)
    assert np.cos(x._val) == cx._val
    for k in x._jacobian.keys():
        assert(cx.partial(k) == -x.partial(k) * np.sin(x._val))

    # Simple Vector Values
    cv = op.cos(v)
    assert ~(np.cos(orig_vals) - cv._val).any()

    # Simple Vector Jacobian
    for key in cv._dict.keys():
        assert ~(cv._dict[key] - v.getDerivative(key) * -np.sin(v._val) * np.eye(len(v._val))).any()

    # Complex Vector Values
    ccv = op.cos(cv)

    assert ~(np.cos(np.cos(orig_vals)) - ccv._val).any()

    # Complex Vector Jacobian
    for key in ccv._dict.keys():
        assert ~(ccv._dict[key] - cv.getDerivative(key) * -np.sin(cv._val) * np.eye(len(cv._val))).any()

    # Constant
    assert op.cos(y) == np.cos(y)


def test_tan():
    # Scalar
    tx = op.tan(x)
    assert np.tan(x._val) == tx._val
    for k in x._jacobian.keys():
        assert tx.partial(k) == x.partial(k) * np.arccos(x._val)**2

    # Simple Vector Values
    tv = op.tan(v)
    assert ~(np.tan(orig_vals) - tv._val).any()

    # Simple Vector Jacobian
    for key in tv._dict.keys():
        assert ~(tv._dict[key] - v.getDerivative(key) * np.arccos(v._val)**2 * np.eye(len(v._val))).any()

    # Complex Vector Values
    tcv = op.tan(tv)

    assert ~(np.tan(np.tan(orig_vals)) - tcv._val).any()

    # Complex Vector Jacobian
    for key in tcv._dict.keys():
        assert ~(tcv._dict[key] - tv.getDerivative(key) * np.arccos(tv._val)**2 * np.eye(len(tv._val))).any()

    # Constant
    assert op.tan(y) == np.tan(y)


def test_arcsin():
    # Scalar
    asx = op.arcsin(x)
    assert np.arcsin(x._val) == asx._val
    for k in x._jacobian.keys():
        assert asx.partial(k) == -x.partial(k) * np.arcsin(x._val) \
               * np.arctan(x._val)

    # Simple Vector Values
    asv = op.arcsin(v)
    assert ~(np.arcsin(orig_vals) - asv._val).any()

    # Simple Vector Jacobian
    for key in asv._dict.keys():
        assert ~(asv._dict[key] - v.getDerivative(key) *-np.arcsin(v._val)*np.arctan(v._val) *
                 np.eye(len(v._val))).any()

    # Complex Vector Values
    ascv = op.arcsin(asv)

    assert ~(np.arcsin(np.arcsin(orig_vals)) - ascv._val).any()

    # Complex Vector Jacobian
    for key in ascv._dict.keys():
        assert ~(ascv._dict[key] - asv.getDerivative(key) *-np.arcsin(asv._val)*np.arctan(asv._val) *
                 np.eye(len(asv._val))).any()

    # Constant
    assert op.arcsin(y) == np.arcsin(y)


def test_arccos():
    # Scalar
    acx = op.arccos(x)
    assert np.arccos(x._val) == acx._val
    for k in x._jacobian.keys():
        assert acx.partial(k) == x.partial(k) * np.arccos(x._val) \
               * np.tan(x._val)

    # Simple Vector Values
    acv = op.arccos(v)
    assert ~(np.arccos(orig_vals) - acv._val).any()

    # Simple Vector Jacobian
    for key in acv._dict.keys():
        assert ~(acv._dict[key] - v.getDerivative(key) * np.arccos(v._val)*np.tan(v._val)*
                 np.eye(len(v._val))).any()

    # Complex Vector Values
    accv = op.arccos(v2)

    assert ~(np.arccos(np.sin(orig_vals)) - accv._val).any()

    # Complex Vector Jacobian
    for key in accv._dict.keys():
        assert ~(accv._dict[key] - v2.getDerivative(key) * np.arccos(v2._val)*np.tan(v2._val) *
                 np.eye(len(v2._val))).any()

    # Constant
    assert op.arccos(y) == np.arccos(y)


def test_arctan():
    # Scalar
    atx = op.arctan(x)
    assert np.arctan(x._val) == atx._val
    for k in x._jacobian.keys():
        assert atx.partial(k) == -x.partial(k) * np.arcsin(x._val)**2

    # Simple Vector Values
    atv = op.arctan(v)
    assert ~(np.arctan(orig_vals) - atv._val).any()

    # Simple Vector Jacobian
    for key in atv._dict.keys():
        assert ~(atv._dict[key] - v.getDerivative(key) * -np.arcsin(v._val)**2 *
                 np.eye(len(v._val))).any()

    # Complex Vector Values
    atcv = op.arctan(v2)

    assert ~(np.arctan(np.sin(orig_vals)) - atcv._val).any()

    # Complex Vector Jacobian
    for key in atcv._dict.keys():
        assert ~(atcv._dict[key] - v2.getDerivative(key) * -np.arcsin(v2._val)**2 * np.eye(len(v2._val))).any()

    # Constant
    assert op.arctan(y) == np.arctan(y)


def test_sinh():
    # Scalar
    shx = op.sinh(x)
    assert np.sinh(x._val) == shx._val
    for k in x._jacobian.keys():
        assert shx.partial(k) == x.partial(k) * np.cosh(x._val)

    # Simple Vector Values
    shv = op.sinh(v)
    assert ~(np.sinh(orig_vals) - shv._val).any()

    # Simple Vector Jacobian
    for key in shv._dict.keys():
        assert ~(shv._dict[key] - v.getDerivative(key) * np.cosh(v._val) *
                 np.eye(len(v._val))).any()

    # Complex Vector Values
    shcv = op.sinh(v2)

    assert ~(np.sinh(np.sin(orig_vals)) - shcv._val).any()

    # Complex Vector Jacobian
    for key in shcv._dict.keys():
        assert ~(shcv._dict[key] - v2.getDerivative(key) * np.cosh(v2._val) * np.eye(len(v2._val))).any()

    # Constant
    assert op.sinh(y) == np.sinh(y)


def test_cosh():
    # Scalar
    chx = op.cosh(x)
    assert np.cosh(x._val) == chx._val
    for k in x._jacobian.keys():
        assert chx.partial(k) == x.partial(k) * np.sinh(x._val)

    # Simple Vector Values
    chv = op.cosh(v)
    assert ~(np.cosh(orig_vals) - chv._val).any()

    # Simple Vector Jacobian
    for key in chv._dict.keys():
        assert ~(chv._dict[key] - v.getDerivative(key) * np.sinh(v._val) *
                 np.eye(len(v._val))).any()

    # Complex Vector Values
    chcv = op.cosh(v2)

    assert ~(np.cosh(np.sin(orig_vals)) - chcv._val).any()

    # Complex Vector Jacobian
    for key in chcv._dict.keys():
        assert ~(chcv._dict[key] - v2.getDerivative(key) * np.sinh(v2._val) * np.eye(len(v2._val))).any()

    # Constant
    assert op.cosh(y) == np.cosh(y)


def test_tanh():
    # Scalar
    thx = op.tanh(x)
    assert np.tanh(x._val) == thx._val
    for k in x._jacobian.keys():
        assert thx.partial(k) == x.partial(k) * (1 - np.tanh(x._val)**2)

    # Simple Vector Values
    thv = op.tanh(v)
    assert ~(np.tanh(orig_vals) - thv._val).any()

    # Simple Vector Jacobian
    for key in thv._dict.keys():
        assert ~(thv._dict[key] - v.getDerivative(key) * (1-np.tanh(v._val)**2) *
                 np.eye(len(v._val))).any()

    # Complex Vector Values
    thcv = op.tanh(v2)

    assert ~(np.tanh(np.sin(orig_vals)) - thcv._val).any()

    # Complex Vector Jacobian
    for key in thcv._dict.keys():
        assert ~(thcv._dict[key] - v2.getDerivative(key) * (1-np.tanh(v2._val)**2) * np.eye(len(v2._val))).any()

    # Constant
    assert op.tanh(y) == np.tanh(y)


def test_arcsinh():
    # Scalar
    ashx = op.arcsinh(x)
    assert np.arcsinh(x._val) == ashx._val
    for k in x._jacobian.keys():
        assert ashx.partial(k) == -x.partial(k) * np.arcsinh(x._val) \
               * np.arctanh(x._val)

    # Simple Vector Values
    ashv = op.arcsinh(v)
    assert ~(np.arcsinh(orig_vals) - ashv._val).any()

    # Simple Vector Jacobian
    for key in ashv._dict.keys():
        assert ~(ashv._dict[key] - v.getDerivative(key) * -np.arcsinh(v._val)*np.arctanh(v._val) *
                 np.eye(len(v._val))).any()

    # Complex Vector Values
    ashcv = op.arcsinh(v2)

    assert ~(np.arcsinh(np.sin(orig_vals)) - ashcv._val).any()

    # Complex Vector Jacobian
    for key in ashcv._dict.keys():
        assert ~(ashcv._dict[key] - v2.getDerivative(key) * -np.arcsinh(v2._val)*np.arctanh(v2._val)
                 * np.eye(len(v2._val))).any()

    # Constant
    assert op.arcsinh(y) == np.arcsinh(y)


def test_arccosh():
    x = ad.create_scalar(1)
    y = 2
    # Scalar
    achx = op.arccosh(x)
    assert np.arccosh(x._val) == achx._val
    for k in x._jacobian.keys():
        assert achx.partial(k) == -x.partial(k) * np.arccosh(x._val) \
               * np.tanh(x._val)

    # Simple Vector Values
    v_temp = ad.create_vector([[2, 3, 4, 5]])[0]
    orig_temp = np.copy(v_temp._val)
    achv = op.arccosh(v_temp)
    assert ~(np.arccosh(orig_temp) - achv._val).any()

    # Simple Vector Jacobian
    for key in achv._dict.keys():
        assert ~(achv._dict[key] - v_temp.getDerivative(key) * -np.arccosh(v_temp._val)*np.tanh(v_temp._val) *
                 np.eye(len(v_temp._val))).any()

    # Complex Vector Values
    achcv = op.arccosh(achv)

    assert ~(np.arccosh(np.arccosh(orig_temp)) - achcv._val).any()

    # Complex Vector Jacobian
    for key in achcv._dict.keys():
        assert ~(achcv._dict[key] - achv.getDerivative(key) * -np.arccosh(achv._val)*np.tanh(achv._val)
                 * np.eye(len(achv._val))).any()

    # Constant
    assert(op.arccosh(y) == np.arccosh(y))


def test_arctanh():
    # Scalar
    athx = op.arctanh(x)
    assert np.arctanh(x._val) == athx._val
    for k in x._jacobian.keys():
        assert athx.partial(k) == x.partial(k) * (1 - np.arctanh(x._val)**2)

    # Simple Vector Values
    athv = op.arctanh(v)
    assert ~(np.arctanh(orig_vals) - athv._val).any()

    # Simple Vector Jacobian
    for key in athv._dict.keys():
        assert ~(athv._dict[key] - v.getDerivative(key) * (1-np.arctanh(v._val)**2) *
                 np.eye(len(v._val))).any()

    # Complex Vector Values
    athcv = op.arctanh(v2)

    assert ~(np.arctanh(np.sin(orig_vals)) - athcv._val).any()

    # Complex Vector Jacobian
    for key in athcv._dict.keys():
        assert ~(athcv._dict[key] - v2.getDerivative(key) * (1-np.arctanh(v2._val)**2)
                 * np.eye(len(v2._val))).any()

    # Constant
    assert op.arctanh(y) == np.arctanh(y)


def test_exp():
    # Scalar
    ex = op.exp(x)
    assert np.exp(x._val) == ex._val
    for k in x._jacobian.keys():
        assert ex.partial(k) == x.partial(k) * np.exp(x._val)

    # Simple Vector Values
    ev = op.exp(v)
    assert ~(np.exp(orig_vals) - ev._val).any()

    # Simple Vector Jacobian
    for key in ev._dict.keys():
        assert ~(ev._dict[key] - v.getDerivative(key) * np.exp(v._val) *
                 np.eye(len(v._val))).any()

    # Complex Vector Values
    ecv = op.exp(v2)

    assert ~(np.exp(np.sin(orig_vals)) - ecv._val).any()

    # Complex Vector Jacobian
    for key in ecv._dict.keys():
        assert ~(ecv._dict[key] - v2.getDerivative(key) * np.exp(v2._val)
                 * np.eye(len(v2._val))).any()

    # Constant
    assert op.exp(y) == np.exp(y)


def test_log():
    # Scalar
    base = 10
    lgn = op.log(n, base)
    assert np.log(n._val) / np.log(base) == lgn._val
    for k in x._jacobian.keys():
        assert lgn.partial(k) == n.partial(k) / (n._val * np.log(base))

    # Simple Vector Values
    lgv = op.log(v, base)
    assert ~(np.log(orig_vals) / np.log(base) - lgv._val).any()

    # Simple Vector Jacobian
    for key in lgv._dict.keys():
        print(lgv._dict[key])
        print(v.getDerivative(key) / (v._val * np.log(base)) *
                 np.eye(len(v._val)))
        assert ~(lgv._dict[key] - v.getDerivative(key) / (v._val * np.log(base)) *
                 np.eye(len(v._val))).any()

    # Complex Vector Values
    lgcv = op.log(v2, base)

    assert ~(np.log(np.sin(orig_vals)) / np.log(base) - lgcv._val).any()

    # Complex Vector Jacobian
    for key in lgcv._dict.keys():
        assert ~(lgcv._dict[key] - v2.getDerivative(key) / (v2._val * np.log(base))
                 * np.eye(len(v2._val))).any()

    # Constant
    z = 24
    assert op.log(z, base) == np.log(z) / np.log(base)


def test_add():
    # Scalar
    res = op.sin(x) + op.sin(z)
    assert np.sin(x._val) + np.sin(z._val) == res._val
    for k in x._jacobian.keys():
        assert res.partial(k) == op.sin(x).partial(k) + op.sin(z).partial(k)

    # Constant
    assert op.sin(y) + op.sin(2 * y) == np.sin(y) + np.sin(2 * y)



from Dotua.nodes.scalar import Scalar
import numpy as np

'''
Initialize local variables for testing. Since these tests need to be
independent of AutoDiff, we will simulate the initalization process by calling
init_jacobian once the entire 'universe' of variables has been defined
'''
# Define scalar objects
vars = x, y = Scalar(1), Scalar(2)
a, b = x.eval(), y.eval()
for var in vars:
    var.init_jacobian(vars)

# Define functions of the scalar objects
f_1 = x + y
f_2 = y + x
f_3 = x - y
f_4 = y - x
f_5 = x * y
f_6 = y * x
f_7 = x / y
f_8 = y / x

# Slightly more complicated functions
g_1 = 10 * x + y / 2 + 1000
g_2 = -2 * x * x - 1 / y

# Exponential functions
h_1 = x ** 2
h_2 = 2 ** x
h_3 = x ** y


def test_jacobian():
    # Test jacobians of scalar primitives
    assert x.partial(x) == 1
    assert x.partial(y) == 0
    assert y.partial(y) == 1
    assert y.partial(x) == 0

    # Test jacobians of functions with addition
    assert f_1.partial(x) == 1
    assert f_1.partial(y) == 1
    assert f_2.partial(x) == 1
    assert f_2.partial(y) == 1

    # Test jacobians of functions with subtraction
    assert f_3.partial(x) == 1
    assert f_3.partial(y) == -1
    assert f_4.partial(x) == -1
    assert f_4.partial(y) == 1

    # Test jacobians of functions with multiplication
    assert f_5.partial(x) == 2
    assert f_5.partial(y) == 1
    assert f_6.partial(x) == 2
    assert f_6.partial(y) == 1

    # Test jacobians of functions with division
    assert f_7.partial(x) == 1/2
    assert f_7.partial(y) == -1/4
    assert f_8.partial(x) == -2
    assert f_8.partial(y) == 1

    # Test jacobians of more complicated functions
    assert g_1.partial(x) == 10
    assert g_1.partial(y) == 1/2
    assert g_2.partial(x) == -4
    assert g_2.partial(y) == 1/4

    # Test jacobians for exponentials and deg > 1 polynomials
    assert h_1.partial(x) == 2*a
    assert h_1.partial(y) == 0
    assert h_2.partial(x) == (2 ** a) * np.log(2)
    assert h_2.partial(y) == 0

def test_add():
    assert f_1.eval() == a + b
    assert f_2.eval() == a + b

    # Directly check commutativity
    assert f_1.eval() == f_2.eval()

    # Check addition with a constant
    radd = 5 + x
    assert radd.eval() == 5 + a


def test_subtract():
    assert f_3.eval() == a - b
    assert f_4.eval() == b - a

    # Check subtraction from a constant
    radd = 5 - x
    assert radd.eval() == 5 - a


def test_multiply():
    assert f_5.eval() == a * b
    assert f_6.eval() == a * b

    # Directly check commutativity
    assert f_5.eval() == f_6.eval()


def test_divide():
    assert f_7.eval() == a / b
    assert f_8.eval() == b / a


def test_power():
    assert h_1.eval() == a ** 2
    assert h_2.eval() == 2 ** a
    assert h_3.eval() == a ** b


def test_other():
    assert g_1.eval() == 10 * a + b / 2 + 1000
    assert g_2.eval() == -2 * (a ** 2) - 1 / b



import numpy as np


class rScalar():
    """
    Allow user defined variables capable of reverse audotmatic differentiation.

    rScalar objects have a single user defined value and 'private' class
    variables parents and grad_val corresponding to functions defined using
    the given rScalar and the current gradient value of the rScalar
    respectively.  The parents of rScalar are defined via operator
    overloading.  Each time the user crreates functions using rScalar
    variables, at least one new rScalar object is created and assigned as the
    parent of the rScalar objects used to create it.
    """

    def __init__(self, val):
        """
        Return an rScalar object with user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        rScalar class instance

        NOTES
        ======
        The val class variable is user defined during initialization.  It can
        be accessed directly but it should never be modified.  The parents and
        grad_val class variables are meant to be 'private' and should never be
        accessed or modified directly by users.  Additionally, note that
        parents is a list of tuples (par, val) where par is an rScalar
        instance and val is the derivative of the rScalar par with respect
        to the rScalar self.
        """
        self._val = val
        self._roots = {}  # keys are input rScalars, vals are lists intemediary rScalars
        self._grad_val = 0

    def _init_roots(self):
        self._roots[self] = [(None, 1)]

    def eval(self):
        """
        Return the value self rScalar object.

        INPUTS
        =======
        self: rScalar class instance

        RETURNS
        =======
        self._val: value of the user defined variable, user defined function,
                   or intermediate rScalar in the computational grpah represented
                   by the self rScalar object

        NOTES
        ======
        rScalar does not overload comparison operators so if users desire to
        compare the values of different rScalar objects they should do so
        by calling eval on each object to obtain the value.
        """
        return self._val

    def gradient(self, input_var):
        """
        Return the derivative of some function with respect to this rScalar.

        INPUTS
        =======
        self: rScalar class instance

        RETURNS
        =======
        self.grad_val: numeric type value repesenting the derivative of a
                       function with respect to this rScalar variable

        NOTES
        ======
        This function is called by rAutoDiff.partial which dictates which
        function's derivative is being calculated with respect to the
        variable represented by the rScalar self.  This method calculates the
        necessary derivative using reverse automatic differentiation by
        recursively finding the derivative of the specified function with
        respect to the given rScalar by defining this derivative in terms of
        the rScalar objects that represent 'intermediary functions' in the
        implicit computational graph that are defined using the rScalar self.
        """
        try:
            for child, val in self._roots[input_var]:
                if child is not None:
                    if child._grad_val == 0:
                        child._grad_val += self._grad_val * val
                    child.gradient(input_var)

        except KeyError:
            raise ValueError("User attempted to differentiate a function " +
                             "respect to a variable on which it is not " +
                             "defined.")

    def __add__(self, other):
        """
        Return an rScalar object whose value is the sum of self and other.

        INPUTS
        =======
        self: rScalar class instance
        other: either a rScalar object or numeric type constant

        RETURNS
        =======
        new_rScalar: a new rScalar object whose value is the sum of the value of
                  the rScalar self and the value of other

        NOTES
        ======
        The rScalar object that is returned by this method is assigned as the
        parent of the rScalar self (and of other if other is an rScalar
        instance).  Specifically, this relationship is stored as a tuple
        (par, val) where par is the new_parent rScalar instance and val is
        the derivative of new_parent with respect to its child rScalar object.
        Storing relationships in this way facilitates the computation of
        gradients through reverse automatic differentiation.
        """
        new_child = rScalar(self._val)
        try:
            new_child._val += other._val
            # For each input variable on which self is defined
            new_child._roots = {input_var: [(self, 1)]
                                for input_var in self._roots.keys()}

            for input_var in other._roots.keys():
                try:
                    new_child._roots[input_var] += [(other, 1)]
                except KeyError:
                    new_child._roots[input_var] = [(other, 1)]
        except AttributeError:
            new_child._val += other
            new_child._roots = {input_var: [(self, 1)]
                                for input_var in self._roots.keys()}
        return new_child

    def __radd__(self, other):
        """Return an rScalar object with value self + other."""
        return self + other

    def __sub__(self, other):
        """Return an rScalar object with value self - other."""
        return self + (-other)

    def __rsub__(self, other):
        """Return an rScalar object with value other - self."""
        return -self + other

    def __mul__(self, other):
        """
        Return an rScalar object whose value is the product of self and other.

        INPUTS
        =======
        self: rScalar class instance
        other: either a rScalar object or numeric type constant

        RETURNS
        =======
        new_rScalar: a new rScalar object whose value is the product of the value
                  of the rScalar self and the value of other

        NOTES
        ======
        The rScalar object that is returned by this method is assigned as the
        parent of the rScalar self (and of other if other is an rScalar
        instance).  Specifically, this relationship is stored as a tuple
        (par, val) where par is the new_parent rScalar instance and val is
        the derivative of new_parent with respect to its child rScalar object.
        Storing relationships in this way facilitates the computation of
        gradients through reverse automatic differentiation.
        """
        new_child = rScalar(self._val)
        try:
            new_child._val *= other._val
            # For each input variable on which self is defined
            new_child._roots = {input_var: [(self, other._val)]
                                for input_var in self._roots.keys()}

            for input_var in other._roots.keys():
                try:
                    new_child._roots[input_var] += [(other, self._val)]
                except KeyError:
                    new_child._roots[input_var] = [(other, self._val)]
        except AttributeError:
            new_child._val *= other
            new_child._roots = {input_var: [(self, other)]
                                for input_var in self._roots.keys()}
        return new_child

    def __rmul__(self, other):
        """Return an rScalar object with value self * other."""
        return self * other

    def __truediv__(self, other):
        """
        Return an rScalar object whose value is the quotient of self and other.

        INPUTS
        =======
        self: rScalar class instance
        other: either a rScalar object or numeric type constant

        RETURNS
        =======
        new_rScalar: a new rScalar object whose value is the quotient of the value
                  of the rScalar self and the value of other

        NOTES
        ======
        The rScalar object that is returned by this method is assigned as the
        parent of the rScalar self (and of other if other is an rScalar
        instance).  Specifically, this relationship is stored as a tuple
        (par, val) where par is the new_parent rScalar instance and val is
        the derivative of new_parent with respect to its child rScalar object.
        Storing relationships in this way facilitates the computation of
        gradients through reverse automatic differentiation.
        """
        new_child = rScalar(self._val)
        try:
            new_child._val /= other._val
            # For each input variable on which self is defined
            new_child._roots = {input_var: [(self, 1 / other._val)]
                                for input_var in self._roots.keys()}

            for input_var in other._roots.keys():
                try:
                    new_child._roots[input_var] += \
                        [(other, -self._val / (other._val ** 2))]
                except KeyError:
                    new_child._roots[input_var] = \
                        [(other, -self._val / (other._val ** 2))]
        except AttributeError:
            new_child._val /= other
            new_child._roots = {input_var: [(self, 1 / other)]
                                for input_var in self._roots.keys()}
        return new_child

    def __rtruediv__(self, other):
        """
        Return an rScalar object whose value is the quotient of other and self.

        INPUTS
        =======
        self: rScalar class instance
        other: either a rScalar object or numeric type constant

        RETURNS
        =======
        new_rScalar: a new rScalar object whose value is the quotient of the value
                  of other nad the value of the rScalar self

        NOTES
        ======
        The rScalar object that is returned by this method is assigned as the
        parent of the rScalar self (and of other if other is an rScalar
        instance).  Specifically, this relationship is stored as a tuple
        (par, val) where par is the new_parent rScalar instance and val is
        the derivative of new_parent with respect to its child rScalar object.
        Storing relationships in this way facilitates the computation of
        gradients through reverse automatic differentiation.   Additionally,
        this method can assume that other is not an instance of rScalar because
        otherwise the division of other and self would be handled by the
        overloading of __truediv__ for the other object.
        """
        new_child = rScalar(other / self._val)
        new_child._roots = {input_var: [(self, -other / (self._val ** 2))]
                            for input_var in self._roots.keys()}
        return new_child


    def __pow__(self, other):
        """
        Return an rScalar object with value self to the power of other.

        INPUTS
        =======
        self: rScalar class instance
        other: either a rScalar object or numeric type constant

        RETURNS
        =======
        new_rScalar: a new rScalar object whose value is the value of the rScalar
                  self raised to the power of the value of other

        NOTES
        ======
        The rScalar object that is returned by this method is assigned as the
        parent of the rScalar self (and of other if other is an rScalar
        instance).  Specifically, this relationship is stored as a tuple
        (par, val) where par is the new_parent rScalar instance and val is
        the derivative of new_parent with respect to its child rScalar object.
        Storing relationships in this way facilitates the computation of
        gradients through reverse automatic differentiation.
        """
        new_child = rScalar(self._val)
        try:
            new_child._val **= other._val
            # For each input variable on which self is defined
            new_child._roots = {input_var: [(self, other._val *
                                self._val ** (other._val - 1))]
                                for input_var in self._roots.keys()}

            for input_var in other._roots.keys():
                try:
                    new_child._roots[input_var] += \
                        [(other, self._val ** other._val * np.log(self._val))]
                except KeyError:
                    new_child._roots[input_var] = \
                        [(other, self._val ** other._val * np.log(self._val))]
        except AttributeError:
            new_child._val **= other
            new_child._roots = {input_var: [(self, other *
                                self._val ** (other - 1))]
                                for input_var in self._roots.keys()}
        return new_child

    def __rpow__(self, other):
        """
        Return an rScalar object with value other to the power of self.

        INPUTS
        =======
        self: rScalar class instance
        other: either a rScalar object or numeric type constant

        RETURNS
        =======
        new_rScalar: a new rScalar object whose value is the value of other raised
                  to the power of the value of the rScalar self

        NOTES
        ======
        The rScalar object that is returned by this method is assigned as the
        parent of the rScalar self (and of other if other is an rScalar
        instance).  Specifically, this relationship is stored as a tuple
        (par, val) where par is the new_parent rScalar instance and val is
        the derivative of new_parent with respect to its child rScalar object.
        Storing relationships in this way facilitates the computation of
        gradients through reverse automatic differentiation.  Additionally,
        this method can assume that other is not an instance of rScalar because
        otherwise the exponentiation of other and self would be handled by the
        overloading of __pow__ for the other object.
        """
        new_child = rScalar(other ** self._val)
        new_child._roots = {input_var:
                            [(self, other ** self._val * np.log(other))]
                            for input_var in self._roots.keys()}
        return new_child

    def __neg__(self):
        """
        Return an rScalar object with the negated value of self.

        INPUTS
        =======
        self: rScalar class instance

        RETURNS
        =======
        A new rScalar object whose value is the negated value of the rScalar
        self

        NOTES
        ======
        The rScalar object that is returned by this method is assigned as the
        parent of the rScalar self.  Specifically, this relationship is stored
        as a tuple (par, val) where par is the new_parent rScalar instance and
        val is the derivative of new_parent with respect to its child rScalar
        object.  Storing relationships in this way facilitates the computation
        of gradients through reverse automatic differentiation.
        """
        new_child = rScalar(-self._val)
        new_child._roots = {input_var: [(self, -1)]
                            for input_var in self._roots.keys()}
        return new_child



from Dotua.rautodiff import rAutoDiff as rad
from Dotua.roperator import rOperator as op
import random

ad = rad()

class NeuralNetwork():
	def __init__(self, input_vals, input_bias, hidden_bias, num_hidden, output, learning_rate = 0.1):
		self.input_vals = input_vals
		self.input_bias = input_bias
		self.hidden_bias = hidden_bias
		self.num_hidden = num_hidden
		self.output = output
		self.learning_rate = learning_rate

		# To initialize the weights from input layer to hidden layer
		self.weights_tohidden = [None] * num_hidden
		for i in range(num_hidden):
			self.weights_tohidden[i] = []
			for j in range(len(input_vals)):
				w = ad.create_rscalar(random.random())
				self.weights_tohidden[i].append(w)

		# To initialize the weights from hidden layer to output layer
		self.weights_tooutput = [None] * len(output)
		for i in range(len(output)):
			self.weights_tooutput[i] = []
			w = ad.create_rscalar(random.random())
			for j in range(num_hidden):
				self.weights_tooutput[i].append(w)

	def train(self, input_vals, output_vals):
		self.input_vals = input_vals
		self.output = output_vals

		# To calculate the hidden layer neurons
		self.hidden_layer = []
		for i in range(self.num_hidden):
			h = 0
			for j in range(len(self.weights_tohidden[i])):
				h = h + self.weights_tohidden[i][j] * self.input_vals[j]
			h = h + self.input_bias
			self.hidden_layer.append(1/(1+op.exp(-h)))

		# To calculate the output layer neurons and error
		error = 0
		for i in range(len(self.output)):
			o = 0
			for j in range(len(self.weights_tooutput[i])):
				o = o + self.weights_tooutput[i][j] * self.hidden_layer[j]
			o = o + self.hidden_bias
			o = 1/(1+op.exp(-o))
			error = error + (o - self.output[i]) ** 2

		# To update weights from hidden layer to output layer
		for i in range(len(self.weights_tooutput)):
			for j in range(len(self.weights_tooutput[i])):
				d = ad.partial(error, self.weights_tooutput[i][j])
				self.weights_tooutput[i][j] = self.weights_tooutput[i][j] - d * self.learning_rate

		# To update weights from input layer to hidden layer
		for i in range(len(self.weights_tohidden)):
			for j in range(len(self.weights_tooutput[i])):
				d = ad.partial(error, self.weights_tohidden[i][j])
				self.weights_tohidden[i][j] = self.weights_tohidden[i][j] - d * self.learning_rate

	def predict(self, input_vals, output_vals):
		self.input_vals = input_vals
		self.output = output_vals

		# To calculate the hidden layer neurons using the current model
		self.hidden_layer = []
		for i in range(self.num_hidden):
			h = 0
			for j in range(len(self.weights_tohidden[i])):
				h = h + self.weights_tohidden[i][j] * self.input_vals[j]
			h = h + self.input_bias
			self.hidden_layer.append(1/(1+op.exp(-h)))

		# To calculate the output layer neurons using the current model and calculate the error
		error = 0
		output_layer = []
		for i in range(len(self.output)):
			o = 0
			for j in range(len(self.weights_tooutput[i])):
				o = o + self.weights_tooutput[i][j] * self.hidden_layer[j]
			o = o + self.hidden_bias
			o = 1/(1+op.exp(-o))
			output_layer.append(o.val)
			error = error + (o - self.output[i]) ** 2
		error = error / len(self.output)
		return (output_layer, error.val)

nn = NeuralNetwork([0.05,0.1],0.35,0.6,2,[0.01,0.09])
for i in range(100):
	nn.train([0.05,0.1],[0.01,0.09])
output, e = nn.predict([0.05,0.1],[0.01,0.09])
print('Final prediction given by the Neural Network is ', output)
print('The mean squared error is ', e)


from Dotua.rautodiff import rAutoDiff as rad
from Dotua.roperator import rOperator as rop


ad = rad()
op = rop()
x,y = ad.create_rvector([[1, 2, 3], [1,3,6]])

a = x + 1
print(ad.partial(a,x))

h = op.sin(x)
print(ad.partial(h,x))

f = x + y
print(ad.partial(f,x))

f = 1 / (x[0] * x[1] * x[2]) + op.sin(1/x[0] + 1/x[1] + 1/x[2])
print(ad.partial(f,x[0]))
print(ad.partial(f,x[1]))
print(ad.partial(f,x[2]))


class Node():
    """
    Abstract interface for computational nodes in automatic differentiation.

    Specifically, this class will provide the abstract interface for the
    Scalar and Vector subclasses.  This class outlines each of the operators
    that must be overloaded by Scalar and Vector.  Users of this AD library
    should NEVER instantiate @Node objects directly (as mentioned in the
    documentation).
    """

    def eval(self):
        raise NotImplementedError

    def __add__(self, other):
        raise NotImplemented

    def __radd__(self, other):
        raise NotImplemented

    def __sub__(self, other):
        raise NotImplemented

    def __rsub__(self, other):
        raise NotImplemented

    def __mul__(self, other):
        raise NotImplemented

    def __rmul__(self, other):
        raise NotImplemented

    def __truediv__(self, other):
        raise NotImplemented

    def __rtruediv__(self, other):
        raise NotImplemented

    def __pow__(self, other):
        raise NotImplemented

    def __rpow__(self, other):
        raise NotImplemented

    def __neg__(self):
        raise NotImplemented



import numpy as np
from Dotua.nodes.rvector import rVector

# Define rVector objects
def generate():
    vars = rVector([10, 10, 15]), rVector([20, 20, 25]), rVector([30, 30, 1])
    for var in vars:
        var._init_roots()
    return tuple(vars)


# Define constants
c_1, c2, c5 = -1, 2, 5


def test_get_item():
    x = rVector([10, 20])
    assert x[0].eval() == 10
    assert x[1].eval() == 20


def test_eval():
    x, y, z = generate()
    assert (x.eval() == x._val).all()
    assert (y.eval() == y._val).all()
    assert (z.eval() == z._val).all()


def test_add():
    x, y, z = generate()
    f = x + y + z
    assert (f.eval() == x._val + y._val + z._val).all()

    f._grad_val = 1
    f.gradient(x)
    assert (x._grad_val == 1).all()
    f.gradient(y)
    assert (y._grad_val == 1).all()
    f.gradient(z)
    assert (z._grad_val == 1).all()

    x, y, z = generate()
    g = z + x + y
    assert (g.eval() == x._val + y._val + z._val).all()

    g._grad_val = 1
    g.gradient(x)
    assert (x._grad_val == 1).all()
    g.gradient(y)
    assert (y._grad_val == 1).all()
    g.gradient(z)
    assert (z._grad_val == 1).all()

    x, y, z = generate()
    h = c_1 + x + y + z + c2
    assert (h.eval() == c_1 + x._val + y._val + z._val + c2).all()

    h._grad_val = 1
    h.gradient(x)
    assert (x._grad_val == 1).all()
    h.gradient(y)
    assert (y._grad_val == 1).all()
    h.gradient(z)
    assert (z._grad_val == 1).all()


def test_subtract():
    x, y, z = generate()
    f = x - y - z
    assert (f.eval() == x._val - y._val - z._val).all()

    f._grad_val = 1
    f.gradient(x)
    assert (x._grad_val == 1).all()
    f.gradient(y)
    assert (y._grad_val == -1).all()
    f.gradient(z)
    assert (z._grad_val == -1).all()

    x, y, z = generate()
    g = z - x - y
    assert (g.eval() == z._val - x._val - y._val).all()

    g._grad_val = 1
    g.gradient(x)
    assert (x._grad_val == -1).all()
    g.gradient(y)
    assert (y._grad_val == -1).all()
    g.gradient(z)
    assert (z._grad_val == 1).all()

    x, y, z = generate()
    h = c_1 - x - y - z - c2
    assert (h.eval() == c_1 - x._val - y._val - z._val - c2).all()

    h._grad_val = 1
    h.gradient(x)
    assert (x._grad_val == -1).all()
    h.gradient(y)
    assert (y._grad_val == -1).all()
    h.gradient(z)
    assert (z._grad_val == -1).all()


def test_multiply():
    x, y, z = generate()
    f = x * y * z
    assert (f.eval() == x._val * y._val * z._val).all()

    f._grad_val = 1
    f.gradient(x)
    assert (x._grad_val == y._val * z._val).all()
    f.gradient(y)
    assert (y._grad_val == x._val * z._val).all()
    f.gradient(z)
    assert (z._grad_val == x._val * y._val).all()

    x, y, z = generate()
    g = z * x * y
    assert (g.eval() == x._val * y._val * z._val).all()

    g._grad_val = 1
    g.gradient(x)
    assert (x._grad_val == y._val * z._val).all()
    g.gradient(y)
    assert (y._grad_val == x._val * z._val).all()
    g.gradient(z)
    assert (z._grad_val == x._val * y._val).all()

    x, y, z = generate()
    h = c_1 * x * y * z * c2
    assert (h.eval() == c_1 * x._val * y._val * z._val * c2).all()

    h._grad_val = 1
    h.gradient(x)
    assert (x._grad_val == c_1 * y._val * z._val * c2).all()
    h.gradient(y)
    assert (y._grad_val == c_1 * x._val * z._val * c2).all()
    h.gradient(z)
    assert (z._grad_val == c_1 * x._val * y._val * c2).all()


def test_divide():
    x, y, z = generate()
    f = x / y / (1 / z)
    assert (f.eval() == x._val / y._val / (1 / z._val)).all()

    f._grad_val = 1
    f.gradient(x)
    assert (x._grad_val == z._val / y._val).all()
    f.gradient(y)
    assert (y._grad_val == -x._val * z._val / y._val ** 2).all()
    f.gradient(z)
    assert (z._grad_val == x._val / y._val).all()

    x, y, z = generate()
    g = z / (x * y)
    assert (g.eval() == z._val / (x._val * y._val)).all()

    g._grad_val = 1
    g.gradient(x)
    assert (x._grad_val == -z._val / (x._val ** 2 * y._val)).all()
    g.gradient(y)
    assert (y._grad_val == -z._val / (x._val * y._val ** 2)).all()
    g.gradient(z)
    assert (z._grad_val == 1 / (x._val * y._val)).all()

    x, y, z = generate()
    h = c_1 * x / (y * z * c2)
    assert (h.eval() == c_1 * x._val / (y._val * z._val * c2)).all()

    h._grad_val = 1
    h.gradient(x)
    assert (x._grad_val == c_1 / (y._val * z._val * c2)).all()
    h.gradient(y)
    assert (y._grad_val ==
            -c_1 * x._val / (y._val * y._val * z._val * c2)).all()
    h.gradient(z)
    assert (z._grad_val == -c_1 * x._val / (y._val * z._val ** 2 * c2)).all()

    x, y, z = generate()
    w = x / c5
    assert (w.eval() == x._val / c5).all()

    w._grad_val = 1
    w.gradient(x)
    assert (x._grad_val == 1 / c5).all()

    exception_raised = False
    try:
        w.gradient(y)
    except ValueError:
        exception_raised = True
    assert exception_raised

    exception_raised = False
    try:
        w.gradient(z)
    except ValueError:
        exception_raised = True
    assert exception_raised


def test_power():
    x, y, z = generate()
    x._val, y._val, z._val = \
        np.array([2, 2, 2]), np.array([3, 3, 3]), np.array([4, 4, 4])
    f = x ** y * z
    assert (f.eval() == x._val ** y._val * z._val).all()

    f._grad_val = 1
    f.gradient(x)
    assert (x._grad_val == y._val * x._val ** (y._val - 1) * z._val).all()
    f.gradient(y)
    assert (y._grad_val == x._val ** y._val * z._val * np.log(x._val)).all()
    f.gradient(z)
    assert (z._grad_val == x._val ** y._val).all()

    x, y, z = generate()
    g = z ** (x * y)
    assert (g.eval() == z._val ** (x._val * y._val)).all()

    g._grad_val = 1
    g.gradient(x)
    assert (x._grad_val == y._val * np.log(z._val) *
            z._val ** (x._val * y._val)).all()
    g.gradient(y)
    assert (y._grad_val == x._val * np.log(z._val) *
            z._val ** (x._val * y._val)).all()
    g.gradient(z)
    assert (z._grad_val == x._val * y._val *
            z._val ** (x._val * y._val - 1)).all()

    x, y, z = generate()
    x._val, y._val, z._val = \
        np.array([2, 2, 2]), np.array([3, 3, 3]), np.array([4, 4, 4])
    h = c_1 * x ** (y ** z * c2)
    assert (h.eval() == c_1 * x._val ** (y._val ** z._val * c2)).all()

    h._grad_val = 1
    h.gradient(x)
    assert (x._grad_val ==
            -c2 * y._val ** z._val * x._val **
            (c2 * y._val ** z._val - 1)).all()
    h.gradient(y)
    assert (y._grad_val == (-c2 * z._val * np.log(x._val) * y._val **
                            (z._val - 1) * x._val ** (c2 *
                            y._val ** z._val))).all()
    h.gradient(z)
    assert (z._grad_val == (-c2 * np.log(x._val) * y._val ** z._val *
                            np.log(y._val) * x._val ** (c2 *
                            y._val ** z._val))).all()

    x, y, z = generate()
    q = 2 ** x
    assert (q.eval() == 2 ** x._val).all()

    q._grad_val = 1
    q.gradient(x)
    assert (x._grad_val == 2 ** x._val * np.log(2)).all()

    exception_raised = False
    try:
        q.gradient(y)
    except ValueError:
        exception_raised = True
    assert exception_raised

    exception_raised = False
    try:
        q.gradient(z)
    except ValueError:
        exception_raised = True
    assert exception_raised

    x, y, z = generate()
    w = x ** 2
    assert (w.eval() == x._val ** 2).all()

    w._grad_val = 1
    w.gradient(x)
    assert (x._grad_val == 2 * x._val).all()

    exception_raised = False
    try:
        w.gradient(y)
    except ValueError:
        exception_raised = True
    assert exception_raised

    exception_raised = False
    try:
        w.gradient(z)
    except ValueError:
        exception_raised = True
    assert exception_raised



# This file serves to test the operator.py module
import math
import numpy as np
from Dotua.nodes.rscalar import rScalar
from Dotua.nodes.rvector import rVector
from Dotua.roperator import rOperator as op


# initializations
def generate(v=0.75):
    var = rScalar(v)
    var._init_roots()
    return var

def generate_vec():
    var = rVector([0.2, 0.3])
    var._init_roots()
    return var


# Define constants for testing
c1, c2 = 0.5, 1


def test_sin():
    # Test rScalar
    x = generate()
    f = op.sin(x)
    assert f.eval() == np.sin(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == np.cos(x._val)

    # Test constant
    assert op.sin(c1) == np.sin(c1)


    # Test vector
    y = generate_vec()
    g = op.sin(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(np.cos(y._val))).all()

def test_cos():
    # Test rScalar
    x = generate()
    f = op.cos(x)
    assert f.eval() == np.cos(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == -np.sin(x._val)

    # Test constant
    assert op.cos(c1) == np.cos(c1)

    # Test vector
    y = generate_vec()
    g = op.cos(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(-np.sin(y._val))).all()



def test_tan():
    # Test rScalar
    x = generate()
    f = op.tan(x)
    assert f.eval() == np.tan(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == np.arccos(x._val) ** 2

    # Test constant
    assert op.tan(c1) == np.tan(c1)

    # Test vector
    y = generate_vec()
    g = op.tan(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(np.arccos(y._val) ** 2)).all()


def test_arcsin():
    # Test rScalar
    x = generate()
    f = op.arcsin(x)
    assert f.eval() == np.arcsin(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == -np.arcsin(x._val) * np.arctan(x._val)

    # Test constant
    assert op.arcsin(c1) == np.arcsin(c1)

    # Test vector
    y = generate_vec()
    g = op.arcsin(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(-np.arcsin(y._val) * np.arctan(y._val))).all()


def test_arccos():
    # Test rScalar
    x = generate(1)
    f = op.arccos(x)
    assert f.eval() == np.arccos(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == np.arccos(x._val) * np.tan(x._val)

    # Test constant
    assert op.arccos(c2) == np.arccos(c2)

    # Test vector
    y = generate_vec()
    g = op.arccos(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(np.arccos(y._val) * np.tan(y._val))).all()


def test_arctan():
    # Test rScalar
    x = generate()
    f = op.arctan(x)
    assert f.eval() == np.arctan(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == -np.arcsin(x._val) ** 2

    # Test constant
    assert op.arctan(c1) == np.arctan(c1)

    # Test vector
    y = generate_vec()
    g = op.arctan(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(-np.arcsin(y._val) ** 2)).all()


def test_sinh():
    x = generate()
    f = op.sinh(x)
    assert f.eval() == np.sinh(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == np.cosh(x._val)

    # Team constnt
    assert op.sinh(c1) == np.sinh(c1)

    # Test vector
    y = generate_vec()
    g = op.sinh(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(np.cosh(y._val))).all()


def test_cosh():
    # Test rScalar
    x = generate()
    f = op.cosh(x)
    assert f.eval() == np.cosh(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == np.sinh(x._val)

    # Test constant
    assert op.cosh(c1) == np.cosh(c1)

    # Test vector
    y = generate_vec()
    g = op.cosh(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(np.sinh(y._val))).all()


def test_tanh():
    x = generate()
    f = op.tanh(x)
    assert f.eval() == np.tanh(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == 1 - np.tanh(x._val) ** 2

    # Test constant
    assert op.tanh(c1) == np.tanh(c1)

    # Test vector
    y = generate_vec()
    g = op.tanh(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(1 - np.tanh(y._val) ** 2)).all()


def test_arcsinh():
    x = generate()
    f = op.arcsinh(x)
    assert f.eval() == np.arcsinh(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == -np.arcsinh(x._val) * np.arctanh(x._val)

    # Test constant
    assert op.arcsinh(c1) == np.arcsinh(c1)

    # Test vector
    y = generate_vec()
    g = op.arcsinh(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(-np.arcsinh(y._val) * np.arctanh(y._val))).all()


def test_arccosh():
    # Test rScalar
    x = generate(1)
    f = op.arccosh(x)
    assert f.eval() == np.arccosh(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == -np.arccosh(x._val) * np.tanh(x._val)

    # Test constant
    assert op.arccosh(c2) == np.arccosh(c2)

    # Test vector
    y = rVector([2, 4])
    y._init_roots()
    g = op.arccosh(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(-np.arccosh(y._val) * np.tanh(y._val))).all()


def test_arctanh():
    # Test rScalar
    x = generate()
    f = op.arctanh(x)
    assert f.eval() == np.arctanh(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == 1 - np.arctanh(x._val) ** 2

    # Test constant
    assert op.arctanh(c1) == np.arctanh(c1)

    # Test vector
    y = generate_vec()
    g = op.arctanh(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(1 - np.arctanh(y._val) ** 2)).all()


def test_exp():
    # Test rScalar
    x = generate()
    f = op.exp(x)
    assert f.eval() == np.exp(x._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == np.exp(x._val)

    # Test constant
    assert op.exp(c1) == np.exp(c1)

    # Test vector
    y = generate_vec()
    g = op.exp(y)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(np.exp(y._val))).all()


def test_log():
    # Test rScalar
    base = 10
    x = generate()
    f = op.log(x, base)
    assert f.eval() == math.log(x._val, base)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == (x._val * math.log(base)) ** (-1)

    # Test constant
    assert op.log(c1, base) == math.log(c1, base)

    # Test vector
    y = generate_vec()
    g = op.log(y, base)
    g._grad_val = 1
    g.gradient(y)
    assert (y._grad_val == np.array(1 / (y._val * np.log(base)))).all()



import math
import numpy as np
from Dotua.nodes.rscalar import rScalar
from Dotua.nodes.rvector import rVector


class rOperator:
    """Returns a new rScalar/rVector object subject to the operator and propagates the
    value and derivative according to reverse mode autodifferentiation.

    The example below pertains to an action on an autodiff rScalar object: x

    Example Usage
    -------------

        $ import rautodiff.rautodiff as rad
        $ x = rad.create_scalar(0)
        $ from rautodiff.roperator import rOperator as rop
        $ y = rop.sin(x)

    """
    @staticmethod
    def sin(x):
        """
        Returns a constant/rScalar/rVector object that is the sine of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the sine function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.sin(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, np.cos(x._val))]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.sin(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, np.cos(x._val))]
                return new_child
            except AttributeError:
                return np.sin(x)

    @staticmethod
    def cos(x):
        """
        Returns a constant/rScalar/rVector object that is the cosine of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the cosine function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.cos(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, -np.sin(x._val))]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.cos(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, -np.sin(x._val))]
                return new_child
            except AttributeError:
                return np.cos(x)

    @staticmethod
    def tan(x):
        """
        Returns a constant/rScalar/rVector object that is the tangent of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the tangent function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.tan(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, np.arccos(x._val)**2)]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.tan(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, np.arccos(x._val)**2)]
                return new_child

            except AttributeError:
                return np.tan(x)

    @staticmethod
    def arcsin(x):
        """
        Returns a constant/rScalar/rVector object that is the arcsine of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the arcsine function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.arcsin(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, -np.arcsin(x._val)*np.arctan(x._val))]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.arcsin(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, -np.arcsin(x._val)*np.arctan(x._val))]
                return new_child

            except AttributeError:
                return np.arcsin(x)

    @staticmethod
    def arccos(x):
        """
        Returns a constant/rScalar/rVector object that is the arccosine of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the arccosine function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.arccos(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, np.arccos(x._val)*np.tan(x._val))]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.arccos(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, np.arccos(x._val)*np.tan(x._val))]
                return new_child

            except AttributeError:
                return np.arccos(x)

    @staticmethod
    def arctan(x):
        """
        Returns a constant/rScalar/rVector object that is the arctan of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the arctan function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.arctan(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, -np.arcsin(x._val)**2)]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.arctan(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, -np.arcsin(x._val)**2)]
                return new_child

            except AttributeError:
                return np.arctan(x)

    @staticmethod
    def sinh(x):
        """
        Returns a constant/rScalar/rVector object that is the sinh of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the sinh function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.sinh(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, np.cosh(x._val))]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.sinh(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, np.cosh(x._val))]
                return new_child

            except AttributeError:
                return np.sinh(x)

    @staticmethod
    def cosh(x):
        """
        Returns a constant/rScalar/rVector object that is the cosh of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the cosh function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.cosh(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, np.sinh(x._val))]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.cosh(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, np.sinh(x._val))]
                return new_child

            except AttributeError:
                return np.cosh(x)

    @staticmethod
    def tanh(x):
        """
        Returns a constant/rScalar/rVector object that is the tanh of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the tanh function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.tanh(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, (1-np.tanh(x._val)**2))]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.tanh(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, (1-np.tanh(x._val)**2))]
                return new_child

            except AttributeError:
                return np.tanh(x)

    @staticmethod
    def arcsinh(x):
        """
        Returns a constant/rScalar/rVector object that is the arcsinh of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the arcsinh function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.arcsinh(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, -np.arcsinh(x._val)*np.arctanh(x._val))]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.arcsinh(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, -np.arcsinh(x._val)*np.arctanh(x._val))]
                return new_child

            except AttributeError:
                return np.arcsinh(x)

    @staticmethod
    def arccosh(x):
        """
        Returns a constant/rScalar/rVector object that is the arccosh of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the arccosh function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.arccosh(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, -np.arccosh(x._val)*np.tanh(x._val))]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.arccosh(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, -np.arccosh(x._val)*np.tanh(x._val))]
                return new_child

            except AttributeError:
                return np.arccosh(x)

    @staticmethod
    def arctanh(x):
        """
        Returns a constant/rScalar/rVector object that is the arctanh of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the arctanh function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.arctanh(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, (1-np.arctanh(x._val)**2))]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.arctanh(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, (1-np.arctanh(x._val)**2))]
                return new_child

            except AttributeError:
                return np.arctanh(x)

    @staticmethod
    def exp(x):
        """
        Returns a constant/rScalar/rVector object that is the exponential of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the exponential function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.exp(x._val))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, np.exp(x._val))]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(np.exp(x._val))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, np.exp(x._val))]
                return new_child

            except AttributeError:
                return np.exp(x)

    @staticmethod
    def log(x, base=np.exp(1)):
        """
        Returns a constant/rScalar/rVector object that is the log of the user specified value.

        INPUTS
        =======
        val: real valued numeric type

        RETURNS
        =======
        constant or rScalar class instance

        NOTES
        ======
        If the input is a constant, a constant is returned with the operation applied. If the input
        is an rScalar object, this method creates a parent node by wrapping the input value in the
        operator function, in this case the log function. The parent is then linked to the child
        for later backpropagation and the parent is returned as a new rScalar object.
        """
        try:
            x._rscalars
            new_child = rVector(np.log(x._val) / np.log(base))
            for input_var in x._roots.keys():
                new_child._roots[input_var] = [(x, 1 / (x._val * math.log(base)))]
            return new_child
        except AttributeError:
            try:
                new_child = rScalar(math.log(x._val, base))
                for input_var in x._roots.keys():
                    new_child._roots[input_var] = [(x, 1 / (x._val * math.log(base)))]
                return new_child

            except AttributeError:
                return math.log(x, base)



import numpy as np
from Dotua.rautodiff import rAutoDiff
from Dotua.roperator import rOperator as op


def test_create_rscalar():
    rad = rAutoDiff()

    # Create a single rScalar
    x = rad.create_rscalar(10)
    assert x.eval() == 10

    # Ensure that roots have been initialized
    assert x in x._roots.keys()

    # Create multiple rScalars
    x, y, z = rad.create_rscalar([1, 3, 6])
    assert x.eval() == 1
    assert y.eval() == 3
    assert z.eval() == 6

    # Ensure that roots have been initialized
    assert x in x._roots.keys()
    assert y in y._roots.keys()
    assert z in z._roots.keys()


def test_create_rvector():
    rad = rAutoDiff()

    # Create a single rScalar
    x = rad.create_rvector([10, 20, 30])
    assert (x.eval() == np.array([10, 20, 30])).all()

    # Ensure that roots have been initialized
    assert x in x._roots.keys()

    # Create multiple rScalars
    x, y, z = rad.create_rvector([[1, 3, 6], [2, 4, 8], [5, 10, 15]])
    assert (x.eval() == np.array([1, 3, 6])).all()
    assert (y.eval() == np.array([2, 4, 8])).all()
    assert (z.eval() == np.array([5, 10, 15])).all()

    # Ensure that roots have been initialized
    assert x in x._roots.keys()
    assert y in y._roots.keys()
    assert z in z._roots.keys()


def test__reset_universe():
    # Test rScalars
    rad = rAutoDiff()
    x, y, z = rad.create_rscalar([1, 3, 6])
    f = x + y - z
    x._grad_val = 1

    rad._reset_universe(f)
    assert x._grad_val == 0

    # Test rVectors
    x, y, z = rad.create_rvector([[1, 3, 6], [2, 4, 8], [5, 10, 15]])
    f = x + y - z
    x._grad_val = 1

    rad._reset_universe(f)
    assert x._grad_val == 0


def test_partial():
    rad = rAutoDiff()

    a, b, c = 0, np.pi / 2, np.pi / 4
    x, y, z = rad.create_rscalar([a, b, c])
    f = x + y - z
    assert rad.partial(f, x) == 1
    assert rad.partial(f, y) == 1
    assert rad.partial(f, z) == -1

    g = x * y
    assert rad.partial(g, x) == b
    assert rad.partial(g, y) == a

    h = op.sin(x * y * z) - op.cos(y)
    assert rad.partial(h, x) == b * c * np.cos(a * b * c)
    assert rad.partial(h, y) == a * c * np.cos(a * b * c) + np.sin(b)
    assert rad.partial(h, z) == a * b * np.cos(a * b * c)

    x, y = rad.create_rvector([[1, 2, 3], [1, 3, 6]])
    h = x + 1
    assert (rad.partial(h, x) == np.array([1, 1, 1])).all()

    f = op.sin(x)
    assert (rad.partial(f, x) == np.cos(np.array([1, 2, 3]))).all()


def test_multiple_partials():
    rad = rAutoDiff()

    a, b, c = 4, 9, 13
    x, y, z = rad.create_rscalar([a, b, c])
    f = x + y - z
    assert rad.partial(f, x) == 1
    assert rad.partial(f, y) == 1
    assert rad.partial(f, z) == -1

    g = x * y + z
    assert rad.partial(g, x) == b
    assert rad.partial(g, y) == a
    assert rad.partial(g, z) == 1

    x, y, z = rad.create_rvector([[1, 3, 6], [2, 4, 8], [5, 10, 15]])
    f = x + y - z
    assert (rad.partial(f, x) == np.array([1, 1, 1])).all()
    assert (rad.partial(f, y) == np.array([1, 1, 1])).all()
    assert (rad.partial(f, z) == np.array([-1, -1, -1])).all()

    g = x * y + z
    assert (rad.partial(g, x) == np.array([2, 4, 8])).all()
    assert (rad.partial(g, y) == np.array([1, 3, 6])).all()
    assert (rad.partial(g, z) == np.array([1, 1, 1])).all()



import numpy as np
from Dotua.nodes.node import Node
from Dotua.nodes.scalar import Scalar


class Vector(Node):
    def __init__(self, val, der = 1):
        """
        Returns a Vector variable with user defined value and derivative

        INPUTS
        =======
        val: list of floats, compulsory
            Value of the Vector variable
        der: float, optional, default value is 1
            Derivative of the Vector variable/function of a variable

        RETURNS
        ========
        Vector class instance

        NOTES
        =====
        PRE:
            - val and der have numeric type and val must be a list
            - two or fewer inputs
        POST:
            returns a Vector class instance with value = val and derivative = der
        """
        self._val = np.array(val)
        self._jacobian = der * np.eye(len(val))
        scalars = [None] * len(val)
        for i in range(len(val)):
            scalars[i] = Scalar(val[i])

            # Initialize the jacobians
        for var in scalars:
            var.init_jacobian(scalars)
        self._scalars = scalars

    def __getitem__(self, idx):
        #return Element(self._val[idx], self._jacobian[idx], self)
        return self._scalars[idx]


    def __add__(self, other):
        """
        Returns the sum of self and other

        INPUTS
        =======
        self: this Vector class instance, compulsory
        other: constant or Vector class instance, compulsory

        RETURNS
        ========
        Vector class instance
        """
        try:
            value = self._val + other._val # If other is a constant, then there will be an attribute error
        except AttributeError:
            value = self._val + other
            try:
                dict_self = self._dict # If self is a user defined variable, then there will be an attribute error
                new = Vector(value, self._jacobian)
                new._dict = dict_self # When self is a complex function and other is a constant, the derivatives of the sum is just the derivatives of self
                return new
            except AttributeError:
                derivative = Counter() # If self is a user defined variable, then we add a dictionary of derivatives of user defined variables to the result Vector variable
                derivative[self] = self._jacobian
                new = Vector(value, self._jacobian)
                new._dict = derivative
                return new
        else:
            try:
                dict_self = self._dict
            except AttributeError:
                dict_self = Counter()
                dict_self[self] = self._jacobian
                try:
                    dict_other = other._dict # If other is a user defined variable, then there will be an attribute error
                except AttributeError:
                    dict_other = Counter()
                    dict_other[other] = other._jacobian  # If other is a user defined variable, then we initiate a Counter dictionary for it
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] + dict_other[key] # Then the derivatives of result Vector variable are sums of derivatives of self and derivatives of other
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
                else:
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] + dict_other[key] # If self and other are both complex functions, then the derivatives of result Vector variable are sums of derivatives of self and derivatives of other
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
            else:
                try:
                    dict_other = other._dict
                except AttributeError:
                    dict_other = Counter()
                    dict_other[other] = other._jacobian
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] + dict_other[key]
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
                else:
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] + dict_other[key]
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new


    def __radd__(self, other):
        """
        Returns the sum of self and other

        INPUTS
        =======
        self: this Vector class instance, compulsory
        other: constant or Vector class instance, compulsory

        RETURNS
        ========
        Vector class instance
        """
        return self + other

    def __sub__(self, other):
        """
        Returns the difference of self and other

        INPUTS
        =======
        self: this Vector class instance, compulsory
        other: constant or Vector class instance, compulsory

        RETURNS
        ========
        Vector class instance
        """
        try:
            value = self._val - other._val # If other is a constant, then there will be an attribute error
        except AttributeError:
            value = self._val - other
            try:
                dict_self = self._dict # If self is a user defined variable, then there will be an attribute
                new = Vector(value, self._jacobian)
                new._dict = dict_self
                return new
            except AttributeError:
                derivative = Counter()
                derivative[self] = self._jacobian
                new = Vector(value, self._jacobian)
                new._dict = derivative # When self is a complex function and other is a constant, the derivatives of the result variable is just the derivatives of self
                return new
        else:
            try:
                dict_self = self._dict
            except AttributeError:
                dict_self = Counter() # If self is a user defined variable, then we add a dictionary of derivatives of user defined variables to the result Vector variable
                dict_self[self] = self._jacobian
                try:
                    dict_other = other._dict # If other is a user defined variable, then there will be an attribute error
                except AttributeError:
                    dict_other = Counter()
                    dict_other[other] = other._jacobian # If other is a user defined variable, then we initiate a Counter dictionary for it
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] - dict_other[key] # Then the derivatives of result Vector variable are differences of derivatives of self and derivatives of other
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
                else:
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] - dict_other[key] # If self and other are both complex functions, then the derivatives of result Vector variable are differences of derivatives of self and derivatives of other
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
            else:
                try:
                    dict_other = other._dict
                except AttributeError:
                    dict_other = Counter()
                    dict_other[other] = other._jacobian
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] - dict_other[key]
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
                else:
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] - dict_other[key]
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new

    def __rsub__(self, other):
        """
        Returns the difference of other and self

        INPUTS
        =======
        self: this Vector class instance, compulsory
        other: constant or Vector class instance, compulsory

        RETURNS
        ========
        Vector class instance
        """
        return self.__neg__() + other

    def __mul__(self, other):
        """
        Returns the product of other and self

        INPUTS
        =======
        self: this Vector class instance, compulsory
        other: constant or Vector class instance, compulsory

        RETURNS
        ========
        Vector class instance
        """
        try:
            val_other = other._val # If other is a constant, then there will be an attribute error
            value = self._val * other._val
        except AttributeError:
            val_other = other
            value = self._val * other
            try:
                dict_self = self._dict # If self is a user defined variable, then there will be an attribute
                for key in dict_self.keys():
                    dict_self[key] = dict_self[key] * val_other
                new = Vector(value, self._jacobian)
                new._dict = dict_self # When self is a complex function and other is a constant, the derivatives of the result variable is just the derivatives of self
                return new
            except AttributeError:
                derivative = Counter() # If self is a user defined variable, then we add a dictionary of derivatives of user defined variables to the result Vector variable
                derivative[self] = self._jacobian * val_other
                new = Vector(value, self._jacobian)
                new._dict = derivative
                return new
        else:
            try:
                dict_self = self._dict
            except AttributeError:
                dict_self = Counter()
                dict_self[self] = self._jacobian
                try:
                    dict_other = other._dict # If other is a user defined variable, then there will be an attribute error
                except AttributeError:
                    dict_other = Counter()
                    dict_other[other] = other._jacobian # If other is a user defined variable, then we initiate a Counter dictionary for it
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] * val_other + dict_other[key] * self._val # Then the derivatives of result Vector variable are sum of products of derivatives and values
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
                else:
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] * val_other + dict_other[key] * self._val # If self and other are both complex functions,then the derivatives of result Vector variable are sum of products of derivatives and values
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
            else:
                try:
                    dict_other = other._dict
                except AttributeError:
                    dict_other = Counter()
                    dict_other[other] = other._jacobian
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] * val_other + dict_other[key] * self._val
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
                else:
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] * val_other + dict_other[key] * self._val
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new

    def __rmul__(self, other):
        """
        Returns the product of other and self

        INPUTS
        =======
        self: this Vector class instance, compulsory
        other: constant or Vector class instance, compulsory

        RETURNS
        ========
        Vector class instance
        """
        return self * other

    def __truediv__(self, other):
        """
        Returns the quotient of self and other

        INPUTS
        =======
        self: this Vector class instance, compulsory
        other: constant or Vector class instance, compulsory

        RETURNS
        ========
        Vector class instance
        """
        try:
            val_other = other._val
            value = self._val / val_other
        except AttributeError:
            val_other = other
            value = self._val / val_other
            try:
                dict_self = self._dict
                for key in dict_self.keys():
                    dict_self[key] = dict_self[key] / val_other
                new = Vector(value, self._jacobian)
                new._dict = dict_self
                return new
            except AttributeError:
                derivative = Counter()
                derivative[self] = self._jacobian / val_other
                new = Vector(value, self._jacobian)
                new._dict = derivative
                return new
        else:
            value = self._val / other._val
            try:
                dict_self = self._dict
            except AttributeError:
                dict_self = Counter()
                dict_self[self] = self._jacobian
                try:
                    dict_other = other._dict
                except AttributeError:
                    dict_other = Counter()
                    dict_other[other] = other._jacobian
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] / val_other - dict_other[key] * self._val / (val_other * val_other)
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
                else:
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] / val_other - dict_other[key] * self._val / (val_other * val_other)
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
            else:
                try:
                    dict_other = other._dict
                except AttributeError:
                    dict_other = Counter()
                    dict_other[other] = other._jacobian
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] / val_other - dict_other[key] * self._val / (val_other * val_other)
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
                else:
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = dict_self[key] / val_other - dict_other[key] * self._val / (val_other * val_other)
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new

    def __rtruediv__(self, other):
        """
        Returns the quotient of other and self

        INPUTS
        =======
        self: this Vector class instance, compulsory
        other: constant or Vector class instance, compulsory

        RETURNS
        ========
        Vector class instance

        NOTES
        ======
        This method can assume that other is not an instance of Vector because
        otherwise the division of other and self would be handled by the
        overloading of __truediv__ for the other object.
        """
        val_other = other
        value = val_other / self._val
        try:
            dict_self = self._dict
            for key in dict_self.keys():
                dict_self[key] = - val_other * dict_self[key] / (self._val * self._val)
            new = Vector(value, self._jacobian)
            new._dict = dict_self
            return new
        except AttributeError:
            derivative = Counter()
            derivative[self] = - val_other * self._jacobian / (self._val * self._val)
            new = Vector(value, self._jacobian)
            new._dict = derivative
            return new

    def __pow__(self, other):
        """
        Returns the other'th power of self

        INPUTS
        =======
        self: this Vector class instance, compulsory
        other: constant or Vector class instance, compulsory

        RETURNS
        ========
        Vector class instance
        """
        try:
            val_other = other._val
            value = np.power(self._val, val_other)
        except AttributeError:
            val_other  = other
            value = np.power(self._val, val_other)
            try:
                dict_self = self._dict
                for key in dict_self.keys():
                    dict_self[key] = dict_self[key] * val_other * self._val ** (val_other - 1)
                new = Vector(value, self._jacobian)
                new._dict = dict_self
                return new
            except AttributeError:
                derivative = Counter()
                derivative[self] = self._jacobian * val_other * self._val ** (val_other - 1)
                new = Vector(value, self._jacobian)
                new._dict = derivative
                return new
        else:
            try:
                dict_self = self._dict
            except AttributeError:
                dict_self = Counter()
                dict_self[self] = self._jacobian
                try:
                    dict_other = other._dict
                except AttributeError:
                    dict_other = Counter()
                    dict_other[other] = other._jacobian
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = self._val ** val_other * (val_other / self._val * dict_self[key] + np.log(self._val) * dict_other[key])
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
                else:
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = self._val ** val_other * (val_other / self._val * dict_self[key] + np.log(self._val) * dict_other[key])
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
            else:
                try:
                    dict_other = other._dict
                except AttributeError:
                    dict_other = Counter()
                    dict_other[other] = other._jacobian
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = self._val ** val_other * (val_other / self._val * dict_self[key] + np.log(self._val) * dict_other[key])
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new
                else:
                    derivative = Counter()
                    lst = list(dict_self.keys()) + list(dict_other.keys())
                    for key in lst:
                        derivative[key] = self._val ** val_other * (val_other / self._val * dict_self[key] + np.log(self._val) * dict_other[key])
                    new = Vector(value, self._jacobian)
                    new._dict = derivative
                    return new

    def __rpow__(self, other):
        """
        Returns the self'th power of other

        INPUTS
        =======
        self: this Vector class instance, compulsory
        other: constant or Vector class instance, compulsory

        RETURNS
        ========
        Vector class instance

        NOTES
        ======
        This method can assume that other is not an instance of Vector because
        otherwise the exponentiation of other and self would be handled by the
        overloading of __pow__ for the other object.
        """
        val_other  = other
        value = np.power(val_other, self._val)
        try:
            dict_self = self._dict
            for key in dict_self.keys():
                dict_self[key] = val_other ** self._val * np.log(val_other) * dict_self[key]
            new = Vector(value, self._jacobian)
            new._dict = dict_self
            return new
        except AttributeError:
            derivative = Counter()
            derivative[self] = val_other ** self._val * np.log(val_other) * self._jacobian
            new = Vector(value, self._jacobian)
            new._dict = derivative
            return new

    def __neg__(self):
        """
        Returns the product of -1 and self

        INPUTS
        =======
        self: this Vector class instance, compulsory

        RETURNS
        ========
        Vector class instance
        """
        value = - self._val
        derivative = - self._jacobian
        new = Vector(value, derivative)
        try:
            dict_self = self._dict
        except AttributeError:
            dict_self = Counter()
            dict_self[self] = derivative
            new._dict = dict_self
            return new
        else:
            for key in dict_self.keys():
                dict_self[key] = - dict_self[key]
            new._dict = dict_self
            return new

    def __repr__(self):
        """
        Returns a description about the Vector variable class instance

        INPUTS
        =======
        self: this Vector class instance, compulsory

        RETURNS
        ========
        description about the Scalar variable class instance: string
        """
        representation = 'Vector variable with value {}'.format(self._val)
        return representation

    def getDerivative(self, x):
        """ Returns the derivative of function self of variable x

        INPUTS
        =======
        self: this Vector class instance, compulsory
        x: user defined variable, compulsory

        RETURNS
        ========
        derivative of function self of user defined variable x: float

        NOTES
        =====
        PRE:
            - x must be user defined variable
        POST:
            - returns a float derivative
        """
        try:
            d = sum(self._dict[x])
        except Exception:
            if self == x:
                return sum(self._jacobian)
            else:
                return 0
        else:
            return d

    def eval(self):
        return list(self._val)

class Counter(dict):
    """
    Data structure for storing derivatives of a function, which is a subclass
    of dict
    """
    
    def __getitem__(self, idx):
        """	It will give 0 if the key is not in the key list of the dictionary.
            So it will give 0 if the called variable is not in the function.
        """
        self.setdefault(idx, 0)
        return dict.__getitem__(self, idx)



from Dotua.nodes.node import Node
import numpy as np


class Scalar(Node):
    """
    Implements the Node interface for single valued, user defined variables.

    Scalar objects have a single user defined value (either defined by user
    initialization or function composition) and a jacobian.  For
    each @Scalar object, the jacobian is a dictionary representation of the
    derivative of the function represented by the @Scalar object with respect
    to all user initialized Scalar objects defined by the same call to
    AutoDiff.create_scalar() that created the Scalar obect(s) used to define
    the function represented by this Scalar object.
    """

    def __init__(self, val, der=None):
        """
        Return a Scalar object with user specified value and derivative.

        INPUTS
        =======
        val: real valued numeric type
        der: dictionary representation of a jacobian

        RETURNS
        =======
        Scalar class instance

        NOTES
        ======
        The _val and _jacobian class variables are defined with leading
        underscores because they should be accessed or modified directly by
        the user.  The user should only access these variables via the
        Scalar object's eval method.
        """
        self._val = val
        self._jacobian = der

    def init_jacobian(self, nodes):
        """
        Initialize the jacobian class variable with appropriate seed values.

        INPUTS
        =======
        self: Scalar class instance
        nodes: list of Scalar objects, created by the same call to
               AutoDiff.create_scalar as the self Scalar object

        RETURNS
        =======
        Nothing is returned by this method.

        NOTES
        ======
        This function is called by AutoDiff.create_scalar after initializing
        all Scalar objects requested by the user.  This ensures that each
        Scalar object that the user might compose to create new Scalar
        functions has a derivative that is well defined with respect to all
        variables in the 'universe'.
        """
        self._jacobian = {node: int(id(self) == id(node)) for node in nodes}

    def eval(self):
        """
        Return the value and derivative of the self Scalar object.

        INPUTS
        =======
        self: Scalar class instance

        RETURNS
        =======
        self._val: value of the user defined variable or function represented
                   by the self Scalar object
        self._jacobian: dictionary representation of the derivative of the self
                        Scalar object (which could be a user defined variable
                        or function) with respect to all variables in the
                        'universe'

        NOTES
        ======
        Scalar does not overload comparison operators so if users desire to
        compare the values of different Scalar objects they should do so
        by calling eval on each object to obtain the value.
        """
        return self._val

    def partial(self, var):
        """Return the derivative of self with respect to var."""
        return self._jacobian[var]

    def __add__(self, other):
        """
        Return a Scalar object whose value is the sum of self and other.

        INPUTS
        =======
        self: Scalar class instance
        other: either a Scalar object or numeric type constant

        RETURNS
        =======
        new_node: a new Scalar object whose value is the sum of the value of
                  the Scalar self and the value of other

        NOTES
        ======
        The Scalar object that is returned by this method has a well defined
        jacobian with respect to all variables in the 'universe' of the Scalar
        object(s) that were used to create this function.
        """
        new_node = Scalar(self._val, self._jacobian)
        try:
            new_node._val += other._val
            new_node._jacobian = {k: (v + other.partial(k))
                                  for k, v in self._jacobian.items()}
        except AttributeError:
            new_node._val += other
        return new_node

    def __radd__(self, other):
        """Return a Scalar object whose value is the sum of self and other."""
        return self + other

    def __sub__(self, other):
        """Return a Scalar object with value self - other."""
        return self + (-other)

    def __rsub__(self, other):
        """Return a Scalar object with value other - self."""
        return -self + other

    def __mul__(self, other):
        """
        Return a Scalar object whose value is the product of self and other.

        INPUTS
        =======
        self: Scalar class instance
        other: either a Scalar object or numeric type constant

        RETURNS
        =======
        new_node: a new Scalar object whose value is the product of the value
                  of the Scalar self and the value of other

        NOTES
        ======
        The Scalar object that is returned by this method has a well defined
        jacobian with respect to all variables in the 'universe' of the Scalar
        object(s) that were used to create this function.
        """
        new_node = Scalar(self._val, self._jacobian)
        try:
            new_node._val *= other._val
            new_node._jacobian = {k: self._val * other.partial(k)
                                  + v * other._val
                                  for k, v in self._jacobian.items()}
        except:
            new_node._val *= other
            new_node._jacobian = \
                {k: v * other for k, v in self._jacobian.items()}
        return new_node

    def __rmul__(self, other):
        """Return a Scalar object with value self * other."""
        return self * other

    def __truediv__(self, other):
        """
        Return a Scalar object whose value is the quotient of self and other.

        INPUTS
        =======
        self: Scalar class instance
        other: either a Scalar object or numeric type constant

        RETURNS
        =======
        new_node: a new Scalar object whose value is the quotient of the value
                  of the Scalar self and the value of other

        NOTES
        ======
        The Scalar object that is returned by this method has a well defined
        jacobian with respect to all variables in the 'universe' of the Scalar
        object(s) that were used to create this function.
        """
        new_node = Scalar(self._val, self._jacobian)
        try:
            new_node._val /= other._val
            new_node._jacobian = \
                {k: (v * other._val - self._val * other.partial(k))
                 / (other._val ** 2) for k, v in self._jacobian.items()}
        except AttributeError:
            new_node._val /= other
            new_node._jacobian = \
                {k: v / other for k, v in self._jacobian.items()}
        return new_node

    def __rtruediv__(self, other):
        """
        Return a Scalar object whose value is quotient of other and self.

        INPUTS
        =======
        self: Scalar class instance
        other: either a Scalar object or numeric type constant

        RETURNS
        =======
        new_node: a new Scalar object whose value is the quotient of the value
                  of other nad the value of the Scalar self

        NOTES
        ======
        The Scalar object that is returned by this method has a well defined
        jacobian with respect to all variables in the 'universe' of the Scalar
        object(s) that were used to create this function.  Additionally, this
        method can assume that other is not an instance of Scalar because
        otherwise the division of other and self would be handled by the
        overloading of __truediv__ for the other object.
        """
        new_node = Scalar(self._val, self._jacobian)
        new_node._val = other / self._val
        new_node._jacobian = \
            {k: other * (-v) / (self._val ** 2)
             for k, v in self._jacobian.items()}
        return new_node

    def __pow__(self, other):
        """
        Return a Scalar object with value self to the power of other.

        INPUTS
        =======
        self: Scalar class instance
        other: either a Scalar object or numeric type constant

        RETURNS
        =======
        new_node: a new Scalar object whose value is the value of the Scalar
                  self raised to the power of the value of other

        NOTES
        ======
        The Scalar object that is returned by this method has a well defined
        jacobian with respect to all variables in the 'universe' of the Scalar
        object(s) that were used to create this function.
        """
        new_node = Scalar(self._val, self._jacobian)
        try:
            new_node._val **= other._val
            new_node._jacobian = \
                {k: (other._val * v / self._val
                 + np.log(self._val) * other.partial(k))
                 * (self._val ** other._val)
                 for k, v in self._jacobian.items()}
        except AttributeError:
            new_node._val **= other
            new_node._jacobian = \
                {k: other * (self._val ** (other - 1)) * v
                 for k, v in self._jacobian.items()}
        return new_node

    def __rpow__(self, other):
        """
        Return a Scalar object with value other to the power of self.

        INPUTS
        =======
        self: Scalar class instance
        other: either a Scalar object or numeric type constant

        RETURNS
        =======
        new_node: a new Scalar object whose value is the value of other raised
                  to the power of the value of the Scalar self

        NOTES
        ======
        The Scalar object that is returned by this method has a well defined
        jacobian with respect to all variables in the 'universe' of the Scalar
        object(s) that were used to create this function.  Additionally, this
        method can assume that other is not an instance of Scalar because
        otherwise the exponentiation of other and self would be handled by the
        overloading of __pow__ for the other object.
        """
        new_node = Scalar(self._val, self._jacobian)
        new_node._val = other ** self._val
        new_node._jacobian = \
            {k: (other ** self._val) * np.log(other) * v
             for k, v in self._jacobian.items()}
        return new_node

    def __neg__(self):
        """
        Return a Scalar object with the negated value of self.

        INPUTS
        =======
        self: Scalar class instance

        RETURNS
        =======
        A new Scalar object whose value is the negated value of the Scalar self

        NOTES
        ======
        The Scalar object that is returned by this method has a well defined
        jacobian with respect to all variables in the 'universe' of the Scalar
        object that was used to create this function.
        """
        jacobian = {k: -1 * v for k, v in self._jacobian.items()}
        return Scalar(-1 * self._val, jacobian)



'''
This file tests that the AutoDiff initializer functions defined in autodiff.py
are correct.
'''
from Dotua.autodiff import AutoDiff


# Test creating a single variable
def test_single_variable():
    x = AutoDiff.create_scalar(5)
    assert x.eval() == 5
    assert x.partial(x) == 1


# Test initializing multiple variables at once
def test_multiple_variables():
    x, y, z = tuple(AutoDiff.create_scalar([1, 2, 3]))
    assert x.eval() == 1
    assert x.partial(x) == 1
    assert x.partial(y) == 0
    assert x.partial(z) == 0

    assert y.eval() == 2
    assert y.partial(x) == 0
    assert y.partial(y) == 1
    assert y.partial(z) == 0

    assert z.eval() == 3
    assert z.partial(x) == 0
    assert z.partial(y) == 0
    assert z.partial(z) == 1



import numpy as np
from Dotua.nodes.rscalar import rScalar
from Dotua.nodes.rvector import rVector


class rAutoDiff():
    def __init__(self):
        self._func = None

    def create_rscalar(self, vals):
        '''
        Return rScalar object(s) with user defined value(s).

        INPUTS
        ======
        vals: list of lists of floats, compulsory
            Value of the list of Vector variables

        RETURNS
        ========
        A list of Vector variables

        NOTES
        =====

        POST:
            returns a list of vector variables with value defined in vals
        '''
        try:
            rscalars = [None] * len(vals)
            for i in range(len(vals)):
                rscalars[i] = rScalar(vals[i])
                rscalars[i]._init_roots()
            return rscalars
        except TypeError:
            rscalar = rScalar(vals)
            rscalar._init_roots()
            return rscalar

    def create_rvector(self, vals):
        '''
        Return rScalar object(s) with user defined value(s).

        INPUTS
        ======
        vals: list of lists of floats, compulsory
            Value of the list of Vector variables

        RETURNS
        ========
        A list of Vector variables

        NOTES
        =====

        POST:
            returns a list of vector variables with value defined in vals
        '''
        try:
            rvectors = [None] * len(vals)
            for i in range(len(vals)):
                rvectors[i] = rVector(vals[i])
                rvectors[i]._init_roots()
            return rvectors
        except TypeError:
            rvector = rVector(vals)
            rvector._init_roots()
            return rvector

    def partial(self, func, var):
        '''
        Returns derivative of the function with regard to the given variable

        INPUTS
        =====
        func: a function of rScalar variables
        var: an rScalar variable

        RETURNS
        =======
        A constant, which is the gradient of func with regarding to var
        '''
        if (self._func != func):
            if self._func is not None:
                self._reset_universe(func)
            self._func = func
        try:
            func._grad_val = np.zeros(len(var._grad_val)) + 1
        except TypeError:
            func._grad_val = 1
        func.gradient(var)
        return var._grad_val

    def _reset_universe(self, func):
        '''
        Reset gradients of nodes in computational graph before next computation

        INPUTS
        =====
        var: user defined input variable (rScalar)
        '''
        try:
            func._grad_val = np.zeros(len(func._grad_val))
        except TypeError:
            func._grad_val = 0
        for val in func._roots.values():
            for child, _ in val:
                if child is not None:
                    self._reset_universe(child)



import setuptools

with open("docs/documentation.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(name='Dotua',
	version='1.0.0',
	description='Package for Forward/Reverse Autodifferentiation',
	author='Nick Stern, Vincent Viego, Summer Yuan, Zach Wehrwein',
	long_description=long_description,
	long_description_content_type="text/markdown",
	packages=setuptools.find_packages(),
	install_requires=[
	  'numpy',
	  'scipy'
	],
	classifiers=[
		"Development Status :: 5 - Production/Stable",
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	url="https://github.com/VV-NS-CY-ZW-CS-207-Organization/cs207-FinalProject",
	license="MIT",
)


from Dotua.nodes.rscalar import rScalar
import numpy as np


'''
Initialize local variables for testing. Since these tests need to be
independent of rAutoDiff, we will simulate calling partial from rAutoDiff by
manually setting the @grad_val of the function we want to compute a partial
for and then redefining the input variables before calling partial for a
different function
'''


# Define rScalar objects
def generate():
    vars = rScalar(10), rScalar(20), rScalar(30)
    for var in vars:
        var._init_roots()
    return tuple(vars)


# Define constants
c_1, c2, c5 = -1, 2, 5


def test_eval():
    x, y, z = generate()
    assert x.eval() == x._val
    assert y.eval() == y._val
    assert z.eval() == z._val


def test_add():
    x, y, z = generate()
    f = x + y + z
    assert f.eval() == x._val + y._val + z._val

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == 1
    f.gradient(y)
    assert y._grad_val == 1
    f.gradient(z)
    assert z._grad_val == 1

    x, y, z = generate()
    g = z + x + y
    assert g.eval() == x._val + y._val + z._val

    g._grad_val = 1
    g.gradient(x)
    assert x._grad_val == 1
    g.gradient(y)
    assert y._grad_val == 1
    g.gradient(z)
    assert z._grad_val == 1

    x, y, z = generate()
    h = c_1 + x + y + z + c2
    assert h.eval() == c_1 + x._val + y._val + z._val + c2

    h._grad_val = 1
    h.gradient(x)
    assert x._grad_val == 1
    h.gradient(y)
    assert y._grad_val == 1
    h.gradient(z)
    assert z._grad_val == 1


def test_subtract():
    x, y, z = generate()
    f = x - y - z
    assert f.eval() == x._val - y._val - z._val

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == 1
    f.gradient(y)
    assert y._grad_val == -1
    f.gradient(z)
    assert z._grad_val == -1

    x, y, z = generate()
    g = z - x - y
    assert g.eval() == z._val - x._val - y._val

    g._grad_val = 1
    g.gradient(x)
    assert x._grad_val == -1
    g.gradient(y)
    assert y._grad_val == -1
    g.gradient(z)
    assert z._grad_val == 1

    x, y, z = generate()
    h = c_1 - x - y - z - c2
    assert h.eval() == c_1 - x._val - y._val - z._val - c2

    h._grad_val = 1
    h.gradient(x)
    assert x._grad_val == -1
    h.gradient(y)
    assert y._grad_val == -1
    h.gradient(z)
    assert z._grad_val == -1


def test_multiply():
    x, y, z = generate()
    f = x * y * z
    assert f.eval() == x._val * y._val * z._val

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == y._val * z._val
    f.gradient(y)
    assert y._grad_val == x._val * z._val
    f.gradient(z)
    assert z._grad_val == x._val * y._val

    x, y, z = generate()
    g = z * x * y
    assert g.eval() == x._val * y._val * z._val

    g._grad_val = 1
    g.gradient(x)
    assert x._grad_val == y._val * z._val
    g.gradient(y)
    assert y._grad_val == x._val * z._val
    g.gradient(z)
    assert z._grad_val == x._val * y._val

    x, y, z = generate()
    h = c_1 * x * y * z * c2
    assert h.eval() == c_1 * x._val * y._val * z._val * c2

    h._grad_val = 1
    h.gradient(x)
    assert x._grad_val == c_1 * y._val * z._val * c2
    h.gradient(y)
    assert y._grad_val == c_1 * x._val * z._val * c2
    h.gradient(z)
    assert z._grad_val == c_1 * x._val * y._val * c2


def test_divide():
    x, y, z = generate()
    f = x / y / (1 / z)
    assert f.eval() == x._val / y._val / (1 / z._val)

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == z._val / y._val
    f.gradient(y)
    assert y._grad_val == -x._val * z._val / y._val ** 2
    f.gradient(z)
    assert z._grad_val == x._val / y._val

    x, y, z = generate()
    g = z / (x * y)
    assert g.eval() == z._val / (x._val * y._val)

    g._grad_val = 1
    g.gradient(x)
    assert x._grad_val == -z._val / (x._val ** 2 * y._val)
    g.gradient(y)
    assert y._grad_val == -z._val / (x._val * y._val ** 2)
    g.gradient(z)
    assert z._grad_val == 1 / (x._val * y._val)

    x, y, z = generate()
    h = c_1 * x / (y * z * c2)
    assert h.eval() == c_1 * x._val / (y._val * z._val * c2)

    h._grad_val = 1
    h.gradient(x)
    assert x._grad_val == c_1 / (y._val * z._val * c2)
    h.gradient(y)
    assert y._grad_val == -c_1 * x._val / (y._val * y._val * z._val * c2)
    h.gradient(z)
    assert z._grad_val == -c_1 * x._val / (y._val * z._val ** 2 * c2)

    x, y, z = generate()
    w = x / c5
    assert w.eval() == x._val / c5

    w._grad_val = 1
    w.gradient(x)
    assert x._grad_val == 1 / c5

    exception_raised = False
    try:
        w.gradient(y)
    except ValueError:
        exception_raised = True
    assert exception_raised

    exception_raised = False
    try:
        w.gradient(z)
    except ValueError:
        exception_raised = True
    assert exception_raised


def test_power():
    x, y, z = generate()
    f = x ** y * z
    assert f.eval() == x._val ** y._val * z._val

    f._grad_val = 1
    f.gradient(x)
    assert x._grad_val == y._val * x._val ** (y._val - 1) * z._val
    f.gradient(y)
    assert y._grad_val == x._val ** y._val * z._val * np.log(x._val)
    f.gradient(z)
    assert z._grad_val == x._val ** y._val

    x, y, z = generate()
    g = z ** (x * y)
    assert g.eval() == z._val ** (x._val * y._val)

    g._grad_val = 1
    g.gradient(x)
    assert x._grad_val == y._val * np.log(z._val) * z._val ** (x._val * y._val)
    g.gradient(y)
    assert y._grad_val == x._val * np.log(z._val) * z._val ** (x._val * y._val)
    g.gradient(z)
    assert z._grad_val == x._val * y._val * z._val ** (x._val * y._val - 1)

    x, y, z = generate()
    x._val, y._val, z._val = 2, 3, 4
    h = c_1 * x ** (y ** z * c2)
    assert h.eval() == c_1 * x._val ** (y._val ** z._val * c2)

    h._grad_val = 1
    h.gradient(x)
    assert x._grad_val == \
        -c2 * y._val ** z._val * x._val ** (c2 * y._val ** z._val - 1)
    h.gradient(y)
    assert y._grad_val == (-c2 * z._val * np.log(x._val) * y._val **
                           (z._val - 1) * x._val ** (c2 * y._val ** z._val))
    h.gradient(z)
    assert z._grad_val == (-c2 * np.log(x._val) * y._val ** z._val *
                           np.log(y._val) * x._val ** (c2 * y._val ** z._val))

    x, y, z = generate()
    q = 2 ** x
    assert q.eval() == 2 ** x._val

    q._grad_val = 1
    q.gradient(x)
    assert x._grad_val == 2 ** x._val * np.log(2)

    exception_raised = False
    try:
        q.gradient(y)
    except ValueError:
        exception_raised = True
    assert exception_raised

    exception_raised = False
    try:
        q.gradient(z)
    except ValueError:
        exception_raised = True
    assert exception_raised

    x, y, z = generate()
    w = x ** 2
    assert w.eval() == x._val ** 2

    w._grad_val = 1
    w.gradient(x)
    assert x._grad_val == 2 * x._val

    exception_raised = False
    try:
        w.gradient(y)
    except ValueError:
        exception_raised = True
    assert exception_raised

    exception_raised = False
    try:
        w.gradient(z)
    except ValueError:
        exception_raised = True
    assert exception_raised



from Dotua.nodes.scalar import Scalar
from Dotua.nodes.vector import Vector


class AutoDiff():
    """
    Provide user interface for creating Scalar and Vector objects.

    The AutoDiff class allows users to define Scalar and Vector objects
    through staticmethods and thus does not need to be instantiated.  In both
    methods, AutoDiff requires that the user create all variables that will be
    used within the same scope with a single call to either create_scalar or
    create_vector.  This requirement allows AutoDiff to create an implicit
    variable 'universe' which facilitates the tracking of function
    derivatives in Scalar and Vector objects.
    """

    @staticmethod
    def create_scalar(vals):
        """
        Return Scalar object(s) with user defined value(s).

        INPUTS
        =======
        vals: a list of numeric types or a single numeric type value

        RETURNS
        =======
        scalar: if vals is a single numeric type, scalar, a Scalar instance
                with user defined value, is returned as a single object
        scalars: if vals is a list of numeric types, scalars, a list of Scalar
                 objects with values corresponding to vals, is returned as a
                 list

        NOTES
        ======
        This method initializes all Scalar objects desired by the user with
        user defined value.  Before returning these Scalar objects, this method
        solidifies the variable 'universe' by seeding the jacobians of each
        of the Scalar objects with appropriate values with respect to all
        Scalars requested by the user.
        """
        try:
            scalars = [None] * len(vals)
            for i in range(len(vals)):
                scalars[i] = Scalar(vals[i])

            # Initialize the jacobians
            for var in scalars:
                var.init_jacobian(scalars)
            return scalars
        except TypeError:
            scalar = Scalar(vals)
            scalar.init_jacobian([scalar])
            return scalar

    @staticmethod
    def create_vector(vals):
        '''
        Return a list of vector variables with values as defined in vals

        INPUTS
        ======
        vals: list of lists of floats, compulsory
            Value of the list of Vector variables

        RETURNS
        ========
        A list of Vector variables

        '''
        vectors = [None] * len(vals)
        for i in range(len(vals)):
            vectors[i] = Vector(vals[i])
        return vectors



def NewtonsMethod(func, x0, tol=1e-15, maxiters=1000):
    '''
    Computes the roots of func through iterative guesses until change is below tolerance.

    Takes in a function (func), a point at which to estimate (x0),
        an error tolerance (tol), and a maximum number of iterations (maxiters).

    func must be composed of 'AutoDiff.Operator' operations and AutoDiff.Scalar structures.

    Example usage:
    '''

    xn = x0
    steps = []

    for i in range(maxiters):

        # Calculate y at this step.
        y = func(xn)._val

        # Calculate derivative at this step
        dy_dx = list(func(xn)._jacobian.values())[0]

        # If y reaches tolerance, stop
        if abs(y) < tol:
            return xn._val, steps
            break

        else:
            steps.append((xn._val, y))
            # Compute Newton Step
            x_next = y / dy_dx

            # Update X
            xn = xn - x_next


    return xn._val, steps
